import uuid

from .landmark import Landmark
from utils.get_time_separation import get_time

class LinkedLandmarks:
    """
    A list of landmarks that are linked together, e.g. in a route.
    Each landmark serves as a node in the linked list, but since we expect these to be consumed through the rest API, a pythonic reference to the next landmark is not well suited. Instead we use the uuid of the next landmark to reference the next landmark in the list. This is not very efficient, but appropriate for the expected use case ("short" trips with onyl few landmarks).
    """
    
    _landmarks = list[Landmark]
    total_time = int
    uuid = str

    def __init__(self, data: list[Landmark] = None) -> None:
        """
        Initialize a new LinkedLandmarks object. This expects an ORDERED list of landmarks, where the first landmark is the starting point and the last landmark is the end point.
        
        Args:
            data (list[Landmark], optional): The list of landmarks that are linked together. Defaults to None.
        """
        self.uuid = uuid.uuid4()
        self._landmarks = data if data else []
        self._link_landmarks()


    def _link_landmarks(self) -> None:
        """
        Create the links between the landmarks in the list by setting their .next_uuid and the .time_to_next attributes.
        """
        self.total_time = 0
        for i, landmark in enumerate(self._landmarks[:-1]):
            landmark.next_uuid = self._landmarks[i + 1].uuid
            time_to_next = get_time(landmark.location, self._landmarks[i + 1].location)
            landmark.time_to_reach_next = time_to_next
            self.total_time += time_to_next

        self._landmarks[-1].next_uuid = None
        self._landmarks[-1].time_to_reach_next = 0


    def __getitem__(self, index: int) -> Landmark:
        return self._landmarks[index]
    

    def __str__(self) -> str:
        return f"LinkedLandmarks, total time: {self.total_time} minutes, {len(self._landmarks)} stops: [\n\t{'\n\t'.join([str(landmark) for landmark in self._landmarks])}\n]"